#define _SVID_SOURCE /* glibc2 needs this */
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdint.h>

#include <sys/timex.h>

int adjtime(struct timeval *itv, struct timeval *otv)
{
  struct timex tx;

  tx.modes = 0;
  if ( itv ) {
    tx.offset = itv->tv_sec * 1000000L + itv->tv_usec;
    tx.modes = ADJ_OFFSET_SINGLESHOT;
  }
  if ( adjtimex(&tx) < 0)
    return -1;
  if ( otv ) {
    otv->tv_sec  = tx.offset / 1000000;
    if ( tx.offset < 0 )
      otv->tv_usec = -(-tx.offset % 1000000);
    else
      otv->tv_usec = tx.offset % 1000000;
  }
  return 0;
}

#define LOG(x...) ERROR(x)

#define ERROR(x...) \
do { \
  printf("%s: ",argv[0]);\
  printf(x); \
  printf("\n"); \
  fprintf(stderr,"%s: ",argv[0]);\
  fprintf(stderr,x);\
  fprintf(stderr,"\n"); \
}while(0)

int main(int argc,char *argv[])
{
  struct timeval tv;
  time_t t;
  int s;

  tv.tv_sec = tv.tv_usec = 0;

  if(argc != 2) {
    ERROR("no new time given !\nUsage: %s newtime",argv[0]);
    return -1;
  } else if(!sscanf(argv[1],"%d",&s) || !(t=(time_t)s)) {
    ERROR("invalid time %s !",argv[1]);
    return -1;
  } else if((tv.tv_sec = (int)(t - time(NULL))) == 0) {
    ERROR("Time diffrence 0 seconds, system time not changed");
    return 0;
  } else if(tv.tv_sec < -7*24*3600 || tv.tv_sec>7*24*3600) {
    ERROR("invalid time (must be within 7 days of current time) !");
    return -1;

  } else if (tv.tv_sec < 30 && tv.tv_sec > -30) {
    if(adjtime(&tv,&tv) < 0) {
      ERROR("can't set time, adjtimex(%lld) failed (%s)!\nTrying stime", (int64_t)tv.tv_sec, strerror(errno));
      if(stime(&t) < 0) { 
	ERROR("can't set time, stime(%lld) failed (%s)!", (int64_t)t, strerror(errno));
      } else { 
	LOG("System time set Ok, method=stime(%lld) (diff = %d / %d sec)!",
	    (int64_t)s, (int)tv.tv_sec, (int)(time(NULL)-t));
	return 0;
      }
    } else {
      LOG("System time set OK, method=adjtimex(%lld) (%d sec)!", 
	  (int64_t)tv.tv_sec, (int)(time(NULL)-t));
      return 0;
    }

  } else if (stime(&t) < 0) 
    ERROR("can't set time, stime(%lld) failed (%s)!", (int64_t)t, strerror(errno));
  else {
    LOG("System time set Ok, method=stime(%lld) (diff %d / %d sec)!", 
	(int64_t)s, (int)tv.tv_sec, (int)(time(NULL)-t));
    return 0;
  }
  return 1;
}
